<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2021 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.2/2065513
 */
namespace Comments\Model;

use Application\Connection\ConnectionFactory;
use Application\Model\AbstractDAO;
use Application\Model\IModelDAO;
use Application\Permissions\IpProtects;
use P4\Connection\ConnectionInterface;
use P4\Model\Fielded\Iterator;

/**
 * Class CommentDAO to handle access to Perforce comment data
 * @package Changes\Model
 */
class CommentDAO extends AbstractDAO
{
    // The Perforce class that handles comments
    const MODEL        = Comment::class;
    const REVIEW_TOPIC = 'reviews';
    const CHANGE_TOPIC = 'changes';

    /**
     * As the fetchAll function in the model takes extra param of protections we have to write
     * this wrapper to handle that.
     * Retrieves all records that match the passed options.
     * Extends parent to compose a search query when fetching by topic.
     *
     * @param array           $options      an optional array of search conditions and/or options
     *                                      supported options are:
     *                                      FETCH_MAXIMUM - set to integer value to limit to the first
     *                                      max' number of entries.
     *                                      FETCH_AFTER - set to an id _after_ which we start collecting
     *                                      FETCH_BY_TOPIC - set to a 'topic' id to limit results
     *                                      FETCH_BY_IDS - provide an array of ids to fetch.
     *                                      not compatible with FETCH_SEARCH or FETCH_AFTER.
     *                                      FETCH_BY_READ_BY - includes comments that have been read by the users
     *                                      FETCH_BY_UNREAD_BY - includes comments that have not been read by the
     *                                      users
     * @param ConnectionInterface|null $connection the connection to set on the filer
     * @param null            $protects            the users protections list.
     * @return  Iterator                           the list of zero or more matching comment objects
     */
    public function fetchAllWithProtection(
        array $options = [],
        ConnectionInterface $connection = null,
        $protects = null
    ) {
        // By default this simply passes on to model, with extra option of protections.
        return call_user_func(
            static::MODEL  . '::fetchAll',
            $options,
            $this->getConnection($connection),
            $protects
        );
    }

    /**
     * Fetch all the comments for a given Review Id.
     * @param String                   $reviewId    the review ID
     * @return mixed
     */
    public function fetchAllReviewComments($reviewId)
    {
        $p4Admin    = $this->services->get(ConnectionFactory::P4_ADMIN);
        $topic      = self::REVIEW_TOPIC.'/'.$reviewId;
        $reviewDAO  = $this->services->get(IModelDAO::REVIEW_DAO);
        $ipProtects = $this->services->get(IpProtects::IP_PROTECTS);
        // To check access.
        $reviewDAO->fetch($reviewId, $p4Admin);

        $options  = [Comment::FETCH_BY_TOPIC => [$topic]];
        $comments = $this->fetchAllWithProtection($options, $p4Admin, $ipProtects);
        return $comments->toArray();
    }

    /**
     * Retrieves the open/closed comment counts stored by p4 index. This can
     * be lower than the actual comment counts due to the potential for race
     * conditions when saving, but should generally be correct.
     *
     * @param  string|array          $topics     one or more topics to count open/closed comments in.
     * @param  ConnectionInterface   $p4         the perforce connection to use
     * @return array        a single array with open/closed count if a string topic is given,
     *                      otherwise an array of open/closed count arrays keyed by topic.
     */
    public function countByTopic($topics, ConnectionInterface $p4)
    {
        $counts = [];
        $topics = array_filter((array)$topics, 'strlen');
        if ($topics) {
            // encode the topics for searching.
            $counts = array_fill_keys($topics, [0, 0]);
            foreach ($topics as $key => $topic) {
                $topics[$key] = Comment::encodeIndexValue($topic);
            }
            $query  = Comment::COUNT_INDEX . '=' . implode('|' . Comment::COUNT_INDEX . '=', $topics);
            $result = $p4->run('search', $query)->getData();
            // search should return one (possibly more) results for each topic
            // results take the form of 'encodedTopic-openedCount-closedCount'
            // take the highest total count we find for each topic.
            foreach ($result as $count) {
                if (!strpos($count, '-')) {
                    continue;
                }
                // pre 2015.2 counts did not include closed comments, so default that to 0
                $parts  = explode('-', $count);
                $topic  = Comment::decodeIndexValue($parts[0]);
                $opened = (int)$parts[1];
                $closed = isset($parts[2]) ? (int)$parts[2] : 0;

                if (isset($counts[$topic]) && array_sum($counts[$topic]) < $opened + $closed) {
                    $counts[$topic] = [$opened, $closed];
                }
            }
        }
        return $counts;
    }
}
