<?php


namespace Reviews\Filter;

use Application\Config\ConfigException;
use Application\Config\ConfigManager;
use Application\Connection\ConnectionFactory;
use Application\Helper\ArrayHelper;
use Application\I18n\TranslatorFactory;
use Application\InputFilter\InputFilter;
use Application\Permissions\ConfigCheck;
use Groups\Model\Group;
use Interop\Container\ContainerInterface;
use Reviews\Model\Review as ReviewModel;
use Reviews\UpdateService;
use Reviews\Validator\Reviewers as ReviewersValidator;
use Laminas\InputFilter\Input;
use Laminas\Http\Request as HttpRequest;

/**
 * Class Participants to filter and validate Participants
 * @package Reviews\Filter
 */
class Participants extends InputFilter implements IParticipants
{
    const COMBINED_REVIEWERS = 'combinedReviewers';
    const REVIEW             = 'review';
    private $review;
    private $translator;
    private $p4Admin;
    private $services;
    private $reviewersValidator;

    /**
     * Participants filter constructor.
     * @param ContainerInterface    $services   application services
     * @param array                 $options
     * @throws \Exception
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        $this->services   = $services;
        $this->p4Admin    = $services->get(ConnectionFactory::P4_ADMIN);
        $this->review     = $options[self::REVIEW];
        $this->translator = $services->get(TranslatorFactory::SERVICE);
        $this->addParticipantsFilter();
    }

    /**
     * It takes the review model and participant data. Remove the blacklist users and then
     * arrange the participant data to correct format so that it can validate using participants
     * filter. Below is the example response:
     * [
     *      "bruno" => [],
     *      "testUser" => [ "required" => 1 ],
     *      "bob" => [ "vote" => [ "value" => 1, "version" => 1, "isStale" => false] ]
     * ]
     * @param ReviewModel $review the review
     * @param array $data Participants data
     * @param string $method request method value
     * @return array|null
     * @throws ConfigException
     * @throws \Exception
     */
    public function getCombinedReviewers($review, $data, $method)
    {
        $data              = $this->removeBlacklistedReviewers($review, $data);
        $combinedReviewers = null;
        $requiredReviewers = isset($data[ReviewModel::REQUIRED_REVIEWERS]) ?
            $data[ReviewModel::REQUIRED_REVIEWERS] : null;
        $reviewers         =
            isset($data[ReviewModel::REVIEWERS])
                ? is_array($data[ReviewModel::REVIEWERS])
                ? $data[ReviewModel::REVIEWERS] : explode(",", $data[ReviewModel::REVIEWERS])
                : null;
        if ($reviewers != null || $requiredReviewers !== null) {
            $combinedReviewers = [];
        }
        if ($reviewers != null) {
            array_map(
                function ($value) use (&$combinedReviewers) {
                    $combinedReviewers[$value] = [];
                },
                array_unique(array_merge($reviewers, (array) $requiredReviewers))
            );
        }
        if ($requiredReviewers !== null) {
            $reviewerQuorums = isset($data[ReviewModel::REVIEWER_QUORUMS]) ? $data[ReviewModel::REVIEWER_QUORUMS] : [];
            array_map(
                function ($key, $value) use (&$combinedReviewers) {
                    $combinedReviewers[$key][IParticipants::REQUIRED] = $value;
                },
                array_keys($reviewerQuorums),
                $reviewerQuorums
            );
            array_map(
                function ($value) use (&$combinedReviewers) {
                    // Groups may already have been set from the quorums field
                    if (!isset($combinedReviewers[$value][IParticipants::REQUIRED])) {
                        $combinedReviewers[$value][IParticipants::REQUIRED] = true;
                    }
                },
                $requiredReviewers
            );
        }
        $projects                 = $review->getProjects();
        $defaultRetainedReviewers = [];
        $defaultRetainedReviewers = UpdateService::mergeDefaultReviewersForProjects(
            $projects,
            $defaultRetainedReviewers,
            $this->p4Admin,
            [UpdateService::ALWAYS_ADD_DEFAULT => false]
        );
        if ($review && $method === HttpRequest::METHOD_POST) {
            // merge new participant data with old participant data
            $combinedReviewers = ArrayHelper::merge($review->getParticipantsData(), $combinedReviewers);
        } elseif ($review  && $method === HttpRequest::METHOD_DELETE) {
            /*Add default retain reviewer to new payload which are not exist in payload and
            removed the existing retained reviewer from the payload. This will help when we ran the reviewers
            validator*/
            $defaultRetainedReviewersKeys = array_keys($defaultRetainedReviewers);
            $combinedReviewersKeys        = array_keys($combinedReviewers);
            sort($defaultRetainedReviewersKeys);
            sort($combinedReviewersKeys);
            if ($defaultRetainedReviewersKeys != $combinedReviewersKeys) {
                foreach ($defaultRetainedReviewers as $key => $value) {
                    if (array_key_exists($key, $combinedReviewers)) {
                        unset($combinedReviewers[$key]);
                    } else {
                        $combinedReviewers[$key] = $value;
                    }
                }
            } else {
                // when all default retain reviewers are added to delete only author need to add
                // This will help when we ran the reviewer validator
                $author = $review->isValidAuthor() ? $review->getAuthorObject()->getId() : null;
                unset($combinedReviewers);
                $combinedReviewers[$author] = [];
            }
        }
        if ($combinedReviewers !== null) {
            if ($review) {
                // Before we update participants we need to preserve any fields already set (for example
                // 'vote' and 'notificationsDisabled' etc)
                foreach ($review->getParticipantsData() as $participant => $participantData) {
                    foreach ($participantData as $participantField => $fieldValue) {
                        // We do not want the old required field as we have worked this out.
                        // We do not want the old minimum required field as we will set this only if retained
                        // participants next
                        if ($method === HttpRequest::METHOD_POST) {
                            if ($participantField !== IParticipants::REQUIRED
                                && $participantField !== ReviewModel::FIELD_MINIMUM_REQUIRED) {
                                $combinedReviewers[$participant][$participantField] = $fieldValue;
                            }
                        }
                        // Preserve FIELD_MINIMUM_REQUIRED from the default retained reviewers if present for all
                        // methods
                        if (isset($defaultRetainedReviewers[$participant]) &&
                            isset($defaultRetainedReviewers[$participant][ReviewModel::FIELD_MINIMUM_REQUIRED])) {
                            $combinedReviewers[$participant][ReviewModel::FIELD_MINIMUM_REQUIRED] =
                                $defaultRetainedReviewers[$participant][ReviewModel::FIELD_MINIMUM_REQUIRED];
                        }
                    }
                }
            }
        }
        return $combinedReviewers;
    }

    /**
     * It checks whether reviewers validator contains the retained reviewers
     * errors.
     * @return boolean
     */
    public function hasRetainedReviewersError()
    {
        return $this->reviewersValidator->hasRetainedReviewersError();
    }

    /**
     * It brings retained reviewers ids if filled by retained reviewer error
     * @return array
     */
    public function hasRetainedReviewers()
    {
        return $this->reviewersValidator->hasRetainedReviewers();
    }

    /**
     * Add filter for validating Participants data
     * @throws \Exception
     */
    private function addParticipantsFilter()
    {
        $review                   = $this->review;
        $projects                 = $review->getProjects();
        $author                   = $review->isValidAuthor() ? $review->getAuthorObject()->getId() : null;
        $input                    = new Input(self::COMBINED_REVIEWERS);
        $this->reviewersValidator = new ReviewersValidator($this->p4Admin, $projects, $author, true);
        $input->getValidatorChain()->attach($this->reviewersValidator);
        $this->add($input);
    }

    /**
     * Removes blacklisted reviewers from the $data object.
     * Note it will only do this if the blacklisted reviewer does not already exist.
     *
     * @param ReviewModel $review the review
     * @param array       $data   values from the request
     * @return array      $data with removing blacklist user and group if any
     * @throws ConfigException
     */
    private function removeBlacklistedReviewers(ReviewModel $review, $data)
    {
        $config          = $this->services->get(ConfigManager::CONFIG);
        $caseSensitive   = $this->services->get(ConnectionFactory::P4_ADMIN)->isCaseSensitive();
        $groupsBlacklist = ConfigManager::getValue($config, ConfigManager::MENTIONS_GROUPS_EXCLUDE_LIST, []);
        $usersBlacklist  = ConfigManager::getValue($config, ConfigManager::MENTIONS_USERS_EXCLUDE_LIST, []);

        if (count($groupsBlacklist) + count($usersBlacklist) == 0) {
            return $data;
        }

        $old      = $review->get();
        $oldNames = $old && isset($old['participantsData']) ? array_keys($old['participantsData']) : [];

        if (isset($data[ReviewModel::REVIEWER])
            && $this->inBlacklist(
                $data[ReviewModel::REVIEWER], $oldNames, $groupsBlacklist, $usersBlacklist, $caseSensitive
            )
        ) {
            unset($data[ReviewModel::REVIEWER]);
        }

        if (isset($data[ReviewModel::REVIEWERS])) {
            foreach ($data[ReviewModel::REVIEWERS] as $index => $reviewer) {
                if ($this->inBlacklist($reviewer, $oldNames, $groupsBlacklist, $usersBlacklist, $caseSensitive)) {
                    unset($data[ReviewModel::REVIEWERS][$index]);
                }
            }
            $data[ReviewModel::REVIEWERS] = array_values($data[ReviewModel::REVIEWERS]);
        }

        if (isset($data[ReviewModel::REQUIRED_REVIEWERS])) {
            foreach ($data[ReviewModel::REQUIRED_REVIEWERS] as $index => $reviewer) {
                if ($this->inBlacklist($reviewer, $oldNames, $groupsBlacklist, $usersBlacklist, $caseSensitive)) {
                    unset($data[ReviewModel::REQUIRED_REVIEWERS][$index]);
                }
            }
            $data[ReviewModel::REQUIRED_REVIEWERS] = array_values($data[ReviewModel::REQUIRED_REVIEWERS]);
        }
        return $data;
    }

    /**
     * Only blacklists if adding a blacklisted user or group that does not already exist in the reviewers.
     *
     * @param string        $name              the name to check
     * @param array         $oldNames          list of users & groups already in the reviewers list
     * @param array         $groupsBlacklist   list of blacklisted groups
     * @param array         $usersBlacklist    list of blacklisted users
     * @param bool          $caseSensitive     if p4d is case sensitive
     * @return bool
     */
    private function inBlacklist($name, $oldNames, $groupsBlacklist, $usersBlacklist, $caseSensitive)
    {
        if (in_array($name, $oldNames)) {
            return false;
        }

        $groupName        = Group::getGroupName($name);
        $groupBlacklisted = Group::isGroupName($name)
            ? ConfigCheck::isBlacklisted($groupName, $groupsBlacklist, $caseSensitive)
            : false;

        if ($groupBlacklisted || ConfigCheck::isBlacklisted($name, $usersBlacklist, $caseSensitive)) {
            return true;
        }

        return false;
    }
}
