<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2021 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.2/2065513
 */
namespace Reviews\Filter;

use Application\Connection\ConnectionFactory;
use Application\Log\SwarmLogger;
use Application\Model\IModelDAO;
use Application\Permissions\PrivateProjects;
use Interop\Container\ContainerInterface;
use Laminas\Filter\AbstractFilter;
use P4\Exception as P4Exception;
use P4\Spec\Exception\NotFoundException as SpecNotFoundException;
use Projects\Model\Project as ProjectModel;

/**
 * Filter to handle 'projects-for-user:<userId>'
 * @package Reviews\Filter
 */
class ProjectsForUser extends AbstractFilter implements IProjectsForUser
{
    private $services;

    /**
     * @inheritDoc
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        $this->services = $services;
    }

    /**
     * If a value is set to a special value 'projects-for-user:<userId>' this function will modify the value
     * to include all of the projects the user is involved in (as determined by $project->isInvolved). For example
     * if 'value' had value 'projects-for-user:swarm' and swarm was involved in projects
     * a, b, and c then 'value' would be changed to be [a, b, c]. This allows a simple parameter to expand into many
     * projects. If the special syntax is not found this filter has no effect
     * @param mixed $value value
     * @return mixed the unchanged value or an array of project ids if the special value is detected. The value will
     * also be returned unchanged if the specified user is not valid.
     * @throws P4Exception
     */
    public function filter($value)
    {
        $retVal = $value;
        if (is_string($value) && preg_match('/^' . self::PROJECTS_FOR_USER . '(.+)$/', $value, $matches)) {
            $p4Admin = $this->services->get(ConnectionFactory::P4_ADMIN);
            $userId  = $matches[1];
            $userDAO = $this->services->get(IModelDAO::USER_DAO);
            try {
                $user       = $userDAO->fetchById($userId, $p4Admin);
                $projectDAO = $this->services->get(IModelDAO::PROJECT_DAO);
                $projects   = $projectDAO->fetchAll([], $p4Admin);
                $groupDAO   = $this->services->get(IModelDAO::GROUP_DAO);
                $allGroups  = $groupDAO->fetchAll([], $p4Admin)->toArray(true);
                $projects   = $this->services->get(PrivateProjects::PROJECTS_FILTER)->filter($projects);
                $myProjects = (array)$projects->filterByCallback(
                    function (ProjectModel $project) use ($user, $allGroups) {
                        return $project->isInvolved($user, null, $allGroups);
                    }
                )->invoke('getId');

                $retVal = $myProjects;
            } catch (SpecNotFoundException $e) {
                $logger = $this->services->get(SwarmLogger::SERVICE);
                $logger->warn(sprintf("User [%s] in [%s] not found", $userId, $value));
            }
        }
        return $retVal;
    }
}
